// -*- mode: objc; tab-width: 4 -*-

#include <string.h>
#include <stdlib.h>

#import  "KAutomataCore.h"

// Keysyms from XK_Hangul_Kiyeog to XK_Hangul_Hieuh include
// consonants not applicable for Cho-seong's, but they are
// not used by this class anyway.
#define IS_CHO(jamokey)  ((jamokey) >= XK_Hangul_Kiyeog && \
						  (jamokey) <= XK_Hangul_Hieuh)
#define IS_JUNG(jamokey) ((jamokey) >= XK_Hangul_A && \
						  (jamokey) <= XK_Hangul_I)
#define IS_JONG(jamokey) ((jamokey) >= XK_Hangul_J_Kiyeog && \
						  (jamokey) <= XK_Hangul_J_Hieuh)
#define MIN(a, b)        ((a) < (b) ? (a) : (b))

typedef int (*BsearchComp)(const void *, const void *);

static char transition_table[13][13] = {
	// START -> { CHO1, CHO2, CHO3, CHO4, JUNG1, JUNG2, JUNG3, JUNG4,
	//            JONG1, JONG2, JONG3, JONG4 }
	{ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	// CHO1 -> { JUNG1, JUNG2, JUNG3, JUNG4, JONG1, JONG2, JONG3, JONG4 }
	{ 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1 },
	// CHO2 -> { CHO4 }
	{ 0, 0, 0, 0, 1, 0 ,0 ,0 ,0 ,0 ,0 ,0 ,0 },
	// CHO3 -> { CHO4, JUNG1, JUNG2, JUNG3, JUNG4, JONG1, JONG2, JONG3, JONG4 }
	{ 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	// CHO4 -> { JUNG1, JUNG2, JUNG3, JUNG4, JONG1, JONG2, JONG3, JONG4 }
	{ 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1 },
	// JUNG1 -> { JONG1, JONG2, JONG3, JONG4 }
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1 },
	// JUNG2 -> { JUNG4 }
	{ 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0 },
	// JUNG3 -> { JUNG4, JONG1, JONG2, JONG3, JONG4 }
	{ 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1 },
	// JUNG4 -> { JONG1, JONG2, JONG3, JONG4 }
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1 },
	// JONG1 -> { }
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	// JONG2 -> { JONG4 }
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	// JONG3 -> { JONG4 }
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 },
	// JONG4 -> { }
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
};

// These jamo values are given using symbols in <X11/keysymdef.h> but
// they have nothing to do with X KeySym.
// (You might make some relationship between the two though.)
static struct dual_jamo_entry {
	JamoKey first;
	JamoKey second;
	JamoKey equivalent;
} dual_jamo_table[] = {
	{ XK_Hangul_O       , XK_Hangul_A       , XK_Hangul_WA   		  },
	{ XK_Hangul_O       , XK_Hangul_AE      , XK_Hangul_WAE  		  },
	{ XK_Hangul_O       , XK_Hangul_I       , XK_Hangul_OE   		  },
	{ XK_Hangul_U       , XK_Hangul_EO      , XK_Hangul_WEO  		  },
	{ XK_Hangul_U       , XK_Hangul_E       , XK_Hangul_WE   		  },
	{ XK_Hangul_U       , XK_Hangul_I       , XK_Hangul_WI   		  },
	{ XK_Hangul_EU      , XK_Hangul_I       , XK_Hangul_YI   		  },
	{ XK_Hangul_J_Kiyeog, XK_Hangul_J_Sios  , XK_Hangul_J_KiyeogSios  },
	{ XK_Hangul_J_Nieun , XK_Hangul_J_Jieuj , XK_Hangul_J_NieunJieuj  },
	{ XK_Hangul_J_Nieun , XK_Hangul_J_Hieuh , XK_Hangul_J_NieunHieuh  },
	{ XK_Hangul_J_Rieul , XK_Hangul_J_Kiyeog, XK_Hangul_J_RieulKiyeog },
	{ XK_Hangul_J_Rieul , XK_Hangul_J_Mieum , XK_Hangul_J_RieulMieum  },
	{ XK_Hangul_J_Rieul , XK_Hangul_J_Pieub , XK_Hangul_J_RieulPieub  },
	{ XK_Hangul_J_Rieul , XK_Hangul_J_Sios  , XK_Hangul_J_RieulSios   },
	{ XK_Hangul_J_Rieul , XK_Hangul_J_Tieut , XK_Hangul_J_RieulTieut  },
	{ XK_Hangul_J_Rieul , XK_Hangul_J_Phieuf, XK_Hangul_J_RieulPhieuf },
	{ XK_Hangul_J_Rieul , XK_Hangul_J_Hieuh , XK_Hangul_J_RieulHieuh  },
	{ XK_Hangul_J_Pieub , XK_Hangul_J_Sios  , XK_Hangul_J_PieubSios	  }
};

static struct dual_jamo_entry double_jamo_table[] = {
	{ XK_Hangul_Kiyeog  , XK_Hangul_Kiyeog  , XK_Hangul_SsangKiyeog   },
	{ XK_Hangul_Dikeud  , XK_Hangul_Dikeud  , XK_Hangul_SsangDikeud   },
	{ XK_Hangul_Pieub   , XK_Hangul_Pieub   , XK_Hangul_SsangPieub    },
	{ XK_Hangul_Sios    , XK_Hangul_Sios    , XK_Hangul_SsangSios     },
	{ XK_Hangul_Jieuj   , XK_Hangul_Jieuj   , XK_Hangul_SsangJieuj    },
	{ XK_Hangul_J_Kiyeog, XK_Hangul_J_Kiyeog, XK_Hangul_J_SsangKiyeog },
	{ XK_Hangul_J_Sios  , XK_Hangul_J_Sios  , XK_Hangul_J_SsangSios	  }
};

static struct {
	JamoKey cho;
	JamoKey jung;
	JamoKey jong;
} johab_code[] = {
	{ 0                    , 0 		      , 0            		   },
	{ XK_VoidSymbol        , 0 		      , XK_VoidSymbol          },
	{ XK_Hangul_Kiyeog     , XK_VoidSymbol, XK_Hangul_J_Kiyeog     },
	{ XK_Hangul_SsangKiyeog, XK_Hangul_A  , XK_Hangul_J_SsangKiyeog},
	{ XK_Hangul_Nieun      , XK_Hangul_AE , XK_Hangul_J_KiyeogSios },
	{ XK_Hangul_Dikeud     , XK_Hangul_YA , XK_Hangul_J_Nieun      },
	{ XK_Hangul_SsangDikeud, XK_Hangul_YAE, XK_Hangul_J_NieunJieuj },
	{ XK_Hangul_Rieul      , XK_Hangul_EO , XK_Hangul_J_NieunHieuh },
	{ XK_Hangul_Mieum      , 0            , XK_Hangul_J_Dikeud     },
	{ XK_Hangul_Pieub      , 0            , XK_Hangul_J_Rieul      },
	{ XK_Hangul_SsangPieub , XK_Hangul_E  , XK_Hangul_J_RieulKiyeog},
	{ XK_Hangul_Sios       , XK_Hangul_YEO, XK_Hangul_J_RieulMieum },
	{ XK_Hangul_SsangSios  , XK_Hangul_YE , XK_Hangul_J_RieulPieub },
	{ XK_Hangul_Ieung      , XK_Hangul_O  , XK_Hangul_J_RieulSios  },
	{ XK_Hangul_Jieuj      , XK_Hangul_WA , XK_Hangul_J_RieulTieut },
	{ XK_Hangul_SsangJieuj , XK_Hangul_WAE, XK_Hangul_J_RieulPhieuf},
	{ XK_Hangul_Cieuc      , 0            , XK_Hangul_J_RieulHieuh },
	{ XK_Hangul_Khieuq     , 0            , XK_Hangul_J_Mieum      },
	{ XK_Hangul_Tieut      , XK_Hangul_OE , 0                      },
	{ XK_Hangul_Phieuf     , XK_Hangul_YO , XK_Hangul_J_Pieub      },
	{ XK_Hangul_Hieuh      , XK_Hangul_U  , XK_Hangul_J_PieubSios  },
	{ 0                    , XK_Hangul_WEO, XK_Hangul_J_Sios       },
	{ 0                    , XK_Hangul_WE , XK_Hangul_J_SsangSios  },
	{ 0                    , XK_Hangul_WI , XK_Hangul_J_Ieung      },
	{ 0                    , 0            , XK_Hangul_J_Jieuj      },
	{ 0                    , 0            , XK_Hangul_J_Cieuc      },
	{ 0                    , XK_Hangul_YU , XK_Hangul_J_Khieuq     },
	{ 0                    , XK_Hangul_EU , XK_Hangul_J_Tieut      },
	{ 0                    , XK_Hangul_YI , XK_Hangul_J_Phieuf     },
	{ 0                    , XK_Hangul_I  , XK_Hangul_J_Hieuh      },
	{ 0                    , 0            , 0                      },
	{ 0                    , 0            , 0                      }
};

static int is_double_jamo(JamoKey jamo);
static int is_dual_jamo(JamoKey jamo);
static int combine_dual_jamo(JamoKey jamo1, JamoKey jamo2);
static int combine_double_jamo(JamoKey jamo1, JamoKey jamo2);
static int combine_double_or_dual_jamo(JamoKey jamo1, JamoKey jamo2);
static int compare_dual_jamo_entries(struct dual_jamo_entry *e1,
									   struct dual_jamo_entry *e2);
static int compare_dual_jamo_equivalents(struct dual_jamo_entry *e1,
										   struct dual_jamo_entry *e2);
static int cand_list_count(JamoKeyAndStateOffset *keyCandidateList);
static char *jamokeys_to_johab(char *johab, JamoKey keys[]);

@implementation KAutomataCore

- init
{
	if ((self = [super init])) {
		[self reset];
		return self;
	}

	[self free];
	return nil;
}

- (int)eatKey:(JamoKeyAndStateOffset *)keyCandidateList
{
	int want_to_jump_to_base;
	int want_to_jump_to;
	JamoKeyAndStateOffset *p;
	int num_cands = cand_list_count(keyCandidateList);
	
	for (p = keyCandidateList; p < keyCandidateList + num_cands; p++) {
		int state_offset;
		
		if (IS_CHO(p->jamo_key))
			want_to_jump_to_base = NCHO1;
		else if (IS_JUNG(p->jamo_key))
			want_to_jump_to_base = NJUNG1;
		else if (IS_JONG(p->jamo_key))
			want_to_jump_to_base = NJONG1;
		else
			return WRONGKEY;

		state_offset = p->state_offset;
		if (state_offset == 4)
			state_offset = 3;
		want_to_jump_to = want_to_jump_to_base + state_offset;

		// test if state -> want_to_jump_to transition is possible.
		if (!transition_table[state][want_to_jump_to])
			continue;			// no link to there
		if (p->state_offset == 4)
			if (is_double_jamo(p->jamo_key)) {
				if (want_to_jump_to == state+2 ||
					want_to_jump_to == state+1)
					// the second key of a double jamo is a double
					// jamo itself
					continue;
			}
			else { // p->jamo_key must be the second key of a double jamo.
				if (num_elected_keys == 0)
					continue; // impossible
				if (elected_keys[num_elected_keys-1] != p->jamo_key)
					continue; // cannot form a double jamo.
			}
		else if (p->state_offset == 3)
			if (is_dual_jamo(p->jamo_key)) {
				if (want_to_jump_to == state+2 ||
					want_to_jump_to == state+1)
					// the second key of a dual jamo is a dual jamo
					// itself.
					continue;
			}
			else { // p->jamo_key must be the second key of a dual jamo.
				if (num_elected_keys == 0)
					continue; // impossible
				if (combine_dual_jamo(elected_keys[num_elected_keys-1],
									  p->jamo_key) < 0)
					continue; // a dual jamo that doesn't exit.
			}
		if (want_to_jump_to >= NJONG1 && want_to_jump_to <= NJONG4 &&
			state >= NCHO1 && state <= NCHO4) // cho -> jong transition
			if (IS_CHO(keyCandidateList[0].jamo_key))
				continue;
		
		if (num_elected_keys >= MAXKEYSPERCHAR)
			return OVERFLOW;

		elected_keys[num_elected_keys] = p->jamo_key;
		memcpy(cand_list_save[num_elected_keys], keyCandidateList,
			   MIN(num_cands+1, MAXCANDIDATES)*sizeof(*keyCandidateList));
		history[num_elected_keys] = state;
		state = want_to_jump_to;
		num_elected_keys++;

		break;
	}

	if (p == keyCandidateList + MAXCANDIDATES || p->jamo_key == 0)
		// no transition was allowed.
		if (IS_JUNG(keyCandidateList[0].jamo_key) &&
			(state == NJONG1 || state == NJONG3 || state == NJONG4) &&
			IS_CHO(cand_list_save[num_elected_keys-1][0].jamo_key))
		{
			int num_elected_keys_save = num_elected_keys;

			elected_keys[num_elected_keys_save - 1] = 0;
			jamokeys_to_johab(composed_char, elected_keys);

			[self reset];
			[self eatKey:cand_list_save[num_elected_keys_save-1]];
			[self eatKey:keyCandidateList];
			return CHARBREAK;
		}
		else {
			strcpy(composed_char, composing_char);
			[[self reset] eatKey:keyCandidateList];
			return CHARBREAK;
		}

	elected_keys[num_elected_keys] = 0;
	jamokeys_to_johab(composing_char, elected_keys);

	return JUSTFINE;
}

- (int)deleteLastKey
{
	if (num_elected_keys > 0) {
		num_elected_keys--;
		elected_keys[num_elected_keys] = 0;
		state = history[num_elected_keys];
		jamokeys_to_johab(composing_char, elected_keys);

		return JUSTFINE;
	}
	else
		return EMPTY;
}

// This method must be called only just after eatKey returns CHARBREAK.
// You may always get some string, but cannot tell if composition is
// complete.
- (char *)getComposedChar
{
	return composed_char;
}

- (char *)getComposingChar
{
	return composing_char;
}

- clear
{
	[self reset];
	strcpy(composed_char, "");
	return self;
}

- reset
{
	state = NSTART;
	num_elected_keys = 0;
	memset(elected_keys, 0, sizeof(elected_keys));
	strcpy(composing_char, "");
	return self;
}

@end

static int is_dual_jamo(JamoKey jamo)
{
	struct dual_jamo_entry k;

	k.equivalent = jamo;

	return bsearch(&k, dual_jamo_table,
				   sizeof(dual_jamo_table)/sizeof(dual_jamo_table[0]),
				   sizeof(dual_jamo_table[0]),
				   (BsearchComp)compare_dual_jamo_equivalents) != NULL;
}

static int is_double_jamo(JamoKey jamo)
{
	struct dual_jamo_entry k;

	k.equivalent = jamo;

	return bsearch(&k, double_jamo_table,
				   sizeof(double_jamo_table)/sizeof(double_jamo_table[0]),
				   sizeof(double_jamo_table[0]),
				   (BsearchComp)compare_dual_jamo_equivalents) != NULL;
}

static int combine_double_or_dual_jamo(JamoKey jamo1, JamoKey jamo2)
{
	if (jamo1 == jamo2)
		return combine_double_jamo(jamo1, jamo2);
	else
		return combine_dual_jamo(jamo1, jamo2);
}

static int combine_dual_jamo(JamoKey jamo1, JamoKey jamo2)
{
	struct dual_jamo_entry *p;
	struct dual_jamo_entry k = { jamo1, jamo2 };
	
	if ((p = bsearch(&k, dual_jamo_table,
					 sizeof(dual_jamo_table)/sizeof(dual_jamo_table[0]),
					 sizeof(dual_jamo_table[0]),
					 (BsearchComp)compare_dual_jamo_entries)) != NULL)
		return p->equivalent;

	return -1;
}

static int combine_double_jamo(JamoKey jamo1, JamoKey jamo2)
{
	struct dual_jamo_entry *p;
	struct dual_jamo_entry k = { jamo1, jamo2 };
	
	if ((p = bsearch(&k, double_jamo_table,
					 sizeof(double_jamo_table)/sizeof(double_jamo_table[0]),
					 sizeof(double_jamo_table[0]),
					 (BsearchComp)compare_dual_jamo_entries)) != NULL)
		return p->equivalent;

	return -1;
}

static int compare_dual_jamo_entries(struct dual_jamo_entry *e1,
									 struct dual_jamo_entry *e2)
{
	if (e1->first != e2->first)
		return e1->first - e2->first;
	if (e1->second != e2->second)
		return e1->second - e2->second;
	return 0;
}

static int compare_dual_jamo_equivalents(struct dual_jamo_entry *e1,
										   struct dual_jamo_entry *e2)
{
	return e1->equivalent - e2->equivalent;
}

static int cand_list_count(JamoKeyAndStateOffset *keyCandidateList)
{
	int n;
	JamoKeyAndStateOffset *p;
	
	for (n = 0, p = keyCandidateList;
		 p < keyCandidateList + MAXCANDIDATES && p->jamo_key != 0;
		 p++)
		n++;

	return n;
}

static char *jamokeys_to_johab(char *dest, JamoKey keys[])
{
	int i = 0;
	JamoKey cho, jung, jong;
	int cho_code, jung_code, jong_code, result;

	if (keys[0] == 0) {			// no jamokeys stored.
		strcpy(dest, "");
		return dest;
	}
	
	// extract cho-seong

	if (IS_CHO(keys[i]))
		if (IS_CHO(keys[i+1])) { // double cho-seong
			cho = combine_double_or_dual_jamo(keys[i], keys[i+1]);
			i += 2;
		}
		else {					// single cho-seong
			cho = keys[i];
			i++;
		}
	else						// no cho-seong
		cho = XK_VoidSymbol;	// filler

	if (IS_JUNG(keys[i]))
		if (IS_JUNG(keys[i+1])) { // double jung-seong
			jung = combine_double_or_dual_jamo(keys[i], keys[i+1]);
			i += 2;
		}
		else {					// single jung-seong
			jung = keys[i];
			i++;
		}
	else
		jung = XK_VoidSymbol;	// filler

	if (IS_JONG(keys[i]))
		if (IS_JONG(keys[i+1]))	// double jong-seong
			jong = combine_double_or_dual_jamo(keys[i], keys[i+1]);
		else					// single jong-seong
			jong = keys[i];
	else
		jong = XK_VoidSymbol;	// filler
	
	for (i = 0; i < 32; i++)
		if (johab_code[i].cho == cho) {
			cho_code = i;
			break;
		}
#ifdef DEBUG
	if (i == 32)
		fprintf(stderr, "Strange cho-seong jamo.\n");
#endif
	for (i = 0; i < 32; i++)
		if (johab_code[i].jung == jung) {
			jung_code = i;
			break;
		}
#ifdef DEBUG
	if (i == 32)
		fprintf(stderr, "Strange jung-seong jamo.\n");
#endif
	for (i = 0; i < 32; i++)
		if (johab_code[i].jong == jong) {
			jong_code = i;
			break;
		}
#ifdef DEBUG
	if (i == 32)
		fprintf(stderr, "Strange jong-seong jamo.\n");
#endif

	result = 0x8000 | (cho_code << 10) | (jung_code << 5) | jong_code;
	dest[0] = result >> 8;
	dest[1] = result & 0xff;

	return dest;
}
